import abc
from typing import TypeVar, Generic
from abc import ABC

Input = TypeVar("Input")
Output = TypeVar("Output")


class Embedding(ABC, Generic[Input, Output]):
    def __init__(self) -> None:
        pass

    @abc.abstractmethod
    def embed(self, data: Input) -> Output:
        pass



import abc
from typing import TypeVar, Generic
from abc import ABC
from embeddings.data.dataset import Dataset

Input = TypeVar("Input")
Output = TypeVar("Output")


class DataLoader(ABC, Generic[Input, Output]):
    @abc.abstractmethod
    def load(self, dataset: Dataset[Input]) -> Output:
        pass



from transformers import AutoTokenizer, AutoModel


def main() -> None:
    tokenizer = AutoTokenizer.from_pretrained("clarin-pl/roberta-polish-kgr10")
    model = AutoModel.from_pretrained("clarin-pl/roberta-polish-kgr10")

    encoded_input = tokenizer.encode("Zażółć gęślą jaźń", return_tensors="pt")
    outputs = model(encoded_input)


if __name__ == "__main__":
    main()



import abc
from typing import TypeVar, Generic
from abc import ABC

Input = TypeVar("Input")
Output = TypeVar("Output")
NewOutput = TypeVar("NewOutput")
OutputInternal = TypeVar("OutputInternal")


class Transformation(ABC, Generic[Input, Output]):
    @abc.abstractmethod
    def transform(self, data: Input) -> Output:
        pass

    def then(
        self, right: "Transformation[Output, NewOutput]"
    ) -> "Transformation[Input, NewOutput]":
        return CombinedTransformations(self, right)


class CombinedTransformations(
    Transformation[Input, Output], Generic[Input, Output, OutputInternal]
):
    def __init__(
        self,
        left: Transformation[Input, OutputInternal],
        right: Transformation[OutputInternal, Output],
    ) -> None:
        self.left = left
        self.right = right

    def transform(self, data: Input) -> Output:
        intermediate = self.left.transform(data)
        return self.right.transform(intermediate)



import abc
from pathlib import Path
from typing import Any, Dict, Optional, List, Tuple

import flair
import numpy as np
from flair.data import Sentence, Corpus, Dictionary
from flair.trainers import ModelTrainer

from embeddings.embedding.flair_embedding import FlairEmbedding
from embeddings.task.task import Task
from embeddings.data.io import T_path
from embeddings.defaults import RESULTS_PATH


class FlairTask(Task[Corpus, Dict[str, np.ndarray]]):
    MODEL_UNDEFINED_EXCEPTION = ValueError("Model undefined. Use build_task_model() first!")

    def __init__(
        self,
        output_path: T_path = RESULTS_PATH,
        task_train_kwargs: Optional[Dict[str, Any]] = None,
    ):
        super().__init__()
        self.model: Optional[flair.nn.Model] = None
        self.output_path: Path = Path(output_path)
        self.task_train_kwargs = task_train_kwargs if task_train_kwargs else {}

    def task(self, data: Corpus) -> Dict[str, np.ndarray]:
        if not self.model:
            raise self.MODEL_UNDEFINED_EXCEPTION

        self.fit(data)
        y_pred, _ = self.predict(data.test)
        y_true = self.get_y(data.test, self.y_type, self.y_dictionary)
        return {"y_true": y_true, "y_pred": y_pred}

    def fit(
        self,
        corpus: Corpus,
    ) -> Dict[Any, Any]:
        self.trainer = ModelTrainer(self.model, corpus)
        log: Dict[Any, Any] = self.trainer.train(
            base_path=self.output_path, **self.task_train_kwargs
        )
        return log

    def predict(self, data: List[Sentence], mini_batch_size: int = 32) -> Tuple[np.ndarray, float]:
        if not self.model:
            raise self.MODEL_UNDEFINED_EXCEPTION

        self.remove_labels_from_data(data, "predicted")

        loss = self.model.predict(
            sentences=data,
            mini_batch_size=mini_batch_size,
            label_name="predicted",
            return_loss=True,
        )

        y_pred = self.get_y(data, y_type="predicted", y_dictionary=self.y_dictionary)
        self.remove_labels_from_data(data, "predicted")
        return y_pred, loss

    @property
    @abc.abstractmethod
    def y_type(self) -> Any:
        pass

    @abc.abstractmethod
    def build_task_model(self, embedding: FlairEmbedding, y_dictionary: Dictionary) -> None:
        pass

    @staticmethod
    @abc.abstractmethod
    def make_y_dictionary(data: Corpus, y_type: Optional[str] = None) -> Dictionary:
        pass

    @property
    @abc.abstractmethod
    def y_dictionary(self) -> Dictionary:
        pass

    @staticmethod
    @abc.abstractmethod
    def get_y(data: List[Sentence], y_type: str, y_dictionary: Dictionary) -> np.ndarray:
        pass

    @staticmethod
    @abc.abstractmethod
    def remove_labels_from_data(data: List[Sentence], y_type: str) -> None:
        pass



import abc
from typing import List, Tuple, Dict, Any

import numpy as np
from datasets import Metric

from embeddings.evaluator.evaluator import Evaluator


class MetricsEvaluator(Evaluator[Dict[str, np.ndarray], Dict[str, Any]]):
    @property
    @abc.abstractmethod
    def metrics(self) -> List[Tuple[Metric, Dict[str, Any]]]:
        pass

    def evaluate(self, data: Dict[str, np.ndarray]) -> Dict[str, Any]:
        return {
            metric.name: metric.compute(
                references=data["y_true"], predictions=data["y_pred"], **kwargs
            )
            for metric, kwargs in self.metrics
        }



from flair.data import Corpus

from embeddings.transformation.transformation import Transformation


class DownsampleFlairCorpusTransformation(Transformation[Corpus, Corpus]):
    def __init__(
        self,
        percentage: float,
        downsample_train: bool = True,
        downsample_dev: bool = True,
        downsample_test: bool = True,
    ):
        self.percentage = percentage
        self.downsample_train = downsample_train
        self.downsample_dev = downsample_dev
        self.downsample_test = downsample_test

    def transform(self, data: Corpus) -> Corpus:
        return data.downsample(
            percentage=self.percentage,
            downsample_train=self.downsample_train,
            downsample_dev=self.downsample_dev,
            downsample_test=self.downsample_test,
        )



import os
from pathlib import Path

EXPERIMENTAL_PATH = Path(os.path.dirname(__file__)).parent.absolute()

RESOURCES_PATH = EXPERIMENTAL_PATH.joinpath("resources")
DATASET_PATH = RESOURCES_PATH.joinpath("datasets")
RESULTS_PATH = RESOURCES_PATH.joinpath("results")



from typing import List, Tuple, Dict, Any

from datasets import Metric, load_metric

from embeddings.evaluator.metrics_evaluator import MetricsEvaluator


class TextClassificationEvaluator(MetricsEvaluator):
    @property
    def metrics(self) -> List[Tuple[Metric, Dict[str, Any]]]:
        return [
            (load_metric("accuracy"), {}),
            (load_metric("f1"), {"average": "macro"}),
            (load_metric("recall"), {"average": "macro"}),
            (load_metric("precision"), {"average": "macro"}),
        ]



from typing import List, Dict, Any

import spacy
from spacy.training import offsets_to_biluo_tags


def convert_spacy_jsonl_to_connl_bilou(
    data: List[Dict[Any, Any]], nlp: spacy.Language, out_path: str
) -> None:
    """Convert spacy jsonl data to connl format."""
    for text in data:
        raw_text = nlp(text["text"])
        spans = text["spans"]
        entities = [(span["start"], span["end"], span["label"]) for span in spans]
        labels = offsets_to_biluo_tags(doc=raw_text, entities=entities)
        tokens = [tok.text for tok in raw_text]

        with open(out_path, "a") as f:
            for t, l in zip(tokens, labels):
                f.write(f"{t} \t {l} \n")
            f.write("\n")



from pathlib import Path
from typing import Optional

import experimental.datasets.utils.misc
import flair
import torch
import typer
from embeddings.defaults import RESULTS_PATH
from flair.embeddings import TransformerWordEmbeddings


app = typer.Typer()


def setup(device: str) -> None:
    flair.device = torch.device(device)  # type:ignore
    flair.set_seed(441)


def evaluate_sequence_tagging(
    embedding: str = typer.Option(..., help="Huggingface embedding model name"),
    dataset_name: str = typer.Option(..., help="Dataset name"),
    dataset_url: Optional[str] = typer.Option(None, help="Dataset download url"),
    root: str = typer.Option(
        RESULTS_PATH.joinpath("tagging"),
        help="Root directory for output files",
    ),
    device: str = typer.Option("cpu", help="Device for Flair e.g. cpu"),
) -> None:
    setup(device)

    out_dir = Path(root, embedding, dataset_name)
    out_dir.mkdir(parents=True, exist_ok=True)

    ds_args = {"url": dataset_url} if dataset_url else {}
    dataset_cls = experimental.datasets.utils.misc.get_dataset_cls(dataset_name)
    dataset = dataset_cls(**ds_args)
    corpus = dataset.to_flair_column_corpus()
    tag_dictionary = corpus.make_tag_dictionary(tag_type="tag")

    test_sentences = corpus.test.sentences.copy()

    print("Loading embeddings...")
    embeddings = TransformerWordEmbeddings(embedding)

    print("Training model...")
    # tagger = FlairSequenceTagger(
    #     embeddings=embeddings,
    #     hidden_dim=256,
    #     tag_dictionary=tag_dictionary,
    #     output_path=str(out_dir),
    # )
    #
    # training_log = tagger.fit(corpus)
    # evaluation_log = tagger.evaluate(test_sentences)
    # with out_dir.joinpath("results.pkl").open(mode="wb") as f:
    #     pickle.dump(obj={"training_log": training_log, "evaluation_log": evaluation_log}, file=f)
    #
    # print("Done!", training_log)


typer.run(evaluate_sequence_tagging)



from typing import Any, Dict, Optional, List

import numpy as np
from flair.data import Sentence, Corpus, Dictionary
from flair.models import SequenceTagger

from embeddings.embedding.flair_embedding import FlairEmbedding
from embeddings.task.flair_task.flair_task import FlairTask
from embeddings.data.io import T_path


class SequenceTagging(FlairTask):
    def __init__(
        self,
        output_path: T_path,
        hidden_size: int,
        task_model_kwargs: Optional[Dict[str, Any]] = None,
        task_train_kwargs: Optional[Dict[str, Any]] = None,
    ) -> None:
        super().__init__(output_path, task_train_kwargs)
        self.hidden_size = hidden_size
        self.task_model_kwargs = task_model_kwargs if task_model_kwargs else {}

    def build_task_model(self, embedding: FlairEmbedding, y_dictionary: Dictionary) -> None:
        self.model = SequenceTagger(
            hidden_size=self.hidden_size,
            embeddings=embedding.model,
            tag_dictionary=y_dictionary,
            tag_type="tag",
            **self.task_model_kwargs
        )

    @property
    def y_type(self) -> Any:
        if self.model:
            return self.model.tag_type
        else:
            raise self.MODEL_UNDEFINED_EXCEPTION

    @staticmethod
    def make_y_dictionary(data: Corpus, y_type: Optional[str] = None) -> Dictionary:
        return data.make_tag_dictionary(y_type)

    @property
    def y_dictionary(self) -> Dictionary:
        if self.model:
            return self.model.tag_dictionary
        else:
            raise self.MODEL_UNDEFINED_EXCEPTION

    @staticmethod
    def get_y(data: List[Sentence], y_type: str, y_dictionary: Dictionary) -> np.ndarray:
        y = []
        for sent in data:
            sent_y = []
            for token in sent:
                sent_y.append(token.get_tag(y_type).value)
            y.append(sent_y)
        return np.array(y)

    @staticmethod
    def remove_labels_from_data(data: List[Sentence], y_type: str) -> None:
        for sent in data:
            for token in sent:
                token.remove_labels(y_type)



from typing import List, Tuple, Dict, Any

from datasets import Metric, load_metric

from embeddings.evaluator.metrics_evaluator import MetricsEvaluator


class SequenceTaggingEvaluator(MetricsEvaluator):
    @property
    def metrics(self) -> List[Tuple[Metric, Dict[str, Any]]]:
        return [
            (load_metric("seqeval"), {}),
        ]



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['embeddings',
 'embeddings.data',
 'embeddings.embedding',
 'embeddings.evaluator',
 'embeddings.model',
 'embeddings.pipeline',
 'embeddings.task',
 'embeddings.task.flair_task',
 'embeddings.transformation',
 'embeddings.transformation.flair_transformation',
 'embeddings.utils',
 'experimental',
 'experimental.datasets',
 'experimental.datasets.utils',
 'experimental.embeddings',
 'experimental.embeddings.language_models',
 'experimental.embeddings.scripts']

package_data = \
{'': ['*']}

install_requires = \
['Pillow>=8.2.0,<9.0.0',
 'datasets>=1.6.1,<2.0.0',
 'flair>=0.8.0,<0.9.0',
 'requests>=2.25.1,<3.0.0',
 'scikit-learn>=0.24.1,<0.25.0',
 'seqeval>=1.2.2,<2.0.0',
 'spacy>=3.0.0,<4.0.0',
 'srsly>=2.4.0,<3.0.0',
 'tensorboard>=2.4.1,<3.0.0',
 'torch>=1.8.0,<2.0.0',
 'transformers>=4.4.2,<5.0.0',
 'typer>=0.3.2,<0.4.0']

setup_kwargs = {
    'name': 'clarinpl-embeddings',
    'version': '0.0.1rc10',
    'description': '',
    'long_description': None,
    'author': 'Roman Bartusiak',
    'author_email': 'riomus@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/CLARIN-PL/embeddings',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



from pathlib import Path
from typing import Optional

import srsly
from embeddings.data.io import T_path
from experimental.datasets.base import BaseDataset
from experimental.datasets.utils.converters import convert_spacy_jsonl_to_connl_bilou
from experimental.datasets.utils.misc import (
    split_train_test_dev,
    DatasetDownloader,
    all_files_exists,
)
from flair.datasets import ColumnCorpus
from spacy.lang import pl


class PromisesElectionsPLDataset(BaseDataset):
    def __init__(
        self,
        url: Optional[str] = None,
        output_path: T_path = "resources/datasets/promises-elections-pl/",
    ):
        super().__init__()
        self.ds_files = ["train.csv", "dev.csv", "test.csv"]
        self.url = url
        self.path = Path(output_path)
        if not all_files_exists(path=self.path, files=self.ds_files):
            if url:
                print("Downloading and preprocessing data.")
                self._preprocess()
            else:
                raise ValueError("Can't download files without download url being set.")

    def _download(self) -> Path:
        assert isinstance(self.url, str)
        downloader = DatasetDownloader(
            root_dir=str(self.path.joinpath("src")),
            url=self.url,
        )
        downloaded = downloader.download()
        if len(downloaded) != 1:
            raise ValueError(f"Download failed. Expected one file, got {len(downloaded)}")

        return downloaded[0]

    def _preprocess(self) -> None:
        path = self._download()
        data = list(srsly.read_jsonl(path))
        splitted_data = split_train_test_dev(data)
        nlp = pl.Language()
        for key, ds in splitted_data.items():
            convert_spacy_jsonl_to_connl_bilou(
                ds, nlp=nlp, out_path=str(self.path.joinpath(f"{key}.csv"))
            )

    def to_flair_column_corpus(self) -> ColumnCorpus:
        return ColumnCorpus(
            data_folder=self.path,
            column_format={0: "text", 1: "tag"},
            train_file="train.csv",
            test_file="test.csv",
            dev_file="dev.csv",
        )



import importlib
import zipfile
from pathlib import Path
from tempfile import NamedTemporaryFile
from typing import List, Any, Dict, Tuple, Optional

import requests
from sklearn.model_selection import train_test_split
from tqdm.auto import tqdm

from embeddings.data.io import T_path


def split_train_test_dev(data: List[Any]) -> Dict[str, List[Any]]:
    train, test = train_test_split(data, test_size=0.4, random_state=1)
    dev, test = train_test_split(test, test_size=0.5, random_state=1)
    return {"train": train, "dev": dev, "test": test}


def get_dataset_cls(name: str, datasets_module: str = "experimental.datasets") -> Any:
    try:
        return getattr(importlib.import_module(datasets_module), name)
    except AttributeError:
        raise NotImplementedError(f"Dataset {name} not supported!")


def download_file(url: str, chunk_size: int = 1024) -> Tuple[Any, str]:
    r = requests.get(url, stream=True)
    if r.status_code != 200:
        raise RuntimeError(
            f"Error while downloading file, response code status: {r.status_code}. Check download url."
        )

    filename = (
        r.headers.get("Content-Disposition", "filename=ds").split("filename=")[1].replace('"', "")
    )
    filesize = int(r.headers.get("Content-Length", "0"))

    pbar = tqdm(total=filesize, unit="iB", unit_scale=True)
    tmp_file = NamedTemporaryFile()

    for data in r.iter_content(chunk_size=chunk_size):
        tmp_file.write(data)
        pbar.update(len(data))

    pbar.close()
    tmp_file.seek(0)
    return tmp_file, filename


def unzip_file(path: str, out_dir: str) -> List[Path]:
    output_dir = Path(out_dir)
    if output_dir.exists():
        raise FileExistsError(f"Given output path {out_dir} exists.")
    zf = zipfile.ZipFile(path, "r")
    zf.extractall(str(output_dir))
    zf.close()
    return list(output_dir.iterdir())


def all_files_exists(path: T_path, files: List[str]) -> bool:
    path = Path(path)
    return all([path.joinpath(file).exists() for file in files])


class DatasetDownloader:
    def __init__(self, root_dir: str, url: str, filename: Optional[str] = None):
        self.root_dir = root_dir
        self.url = url
        self.filename = filename

    def download(self) -> List[Path]:
        tmp_dl_file, filename = download_file(self.url)
        if not self.filename:
            self.filename = filename

        output_paths = []
        if zipfile.is_zipfile(tmp_dl_file.name):
            output_paths.extend(unzip_file(path=tmp_dl_file.name, out_dir=self.root_dir))
        else:
            root_dir = Path(self.root_dir)
            root_dir.mkdir(parents=True)
            output_paths.append(root_dir.joinpath(self.filename))
            with output_paths[-1].open(mode="wb") as f:
                f.write(tmp_dl_file.read())

        tmp_dl_file.close()
        return output_paths
